"""
Class for handling GROUPS_SEGMENT in meta file.
"""
import re
import io
from importlib.resources import files

from sis_meta.groups._parse import parse_groups_segment
from sis_meta.groups._serialize import serialize_groups_segment

class GroupsSegment:
    """
    A class for handling GROUPS_SEGMENT.
    """
    def __init__(self, raw_data=None):
        self.data = None

        if raw_data is None:
            path = files('sis_meta.groups.templates') / 'grouptree.meta'
            with open(path, 'rb') as raw_file:
                raw_data = raw_file.read()
                self.data = parse_groups_segment(raw_data)
        else:
            self.data = parse_groups_segment(raw_data)
        self.group_id = self._find_max_id(self.data)

    def __repr__(self):
        return serialize_groups_segment(self.data)

    def _find_max_id(self, data):
        """
        Find the higher ID in GROUPS_SEGMENT.
        """
        for dict_ in data:
            max_value = dict_['id']
            if 'child' in dict_:
                current_value = self._find_max_id(dict_['child'])
                max_value = current_value if current_value > max_value else max_value
        return max_value

    def _find_group(self, data, path):
        """
        Find a GROUP dict.

        Parameters
        ----------
        data : list of dict
            An attribute of the :class:`sis_meta.groups_segment.GroupsSegment`.
        path : list
            list of GROUP items.

        Returns
        -------
        dict or None
            A GROUP dict. If the path does not exist, then return `None`
        """
        group_name = path[0]
        group_name_path = path[1:]

        # Base case
        if len(group_name_path) == 0: # Each element in the path has been consumed
            for dict_ in data:
                if dict_['name'] == group_name:
                    return dict_
            return None

        # Recursive case
        for dict_ in data:
            if not dict_['name'] == group_name:
                continue

            if 'child' in dict_:
                return self._find_group(dict_['child'], group_name_path)

    def _find_group_by_id(self, data, id_):
        """
        Find a GROUP dict.

        Parameters
        ----------
        data : list of dict
            An attribute of the :class:`sis_meta.groups_segment.GroupsSegment`.
        id_ : int
            The group id.

        Returns
        -------
        dict
            A GROUP dict.
        """
        # Base case
        for dict_ in data:
            # Base case
            if dict_['id'] == id_:
                return dict_

            # Recursive case
            if 'child' in dict_:
                result = self._find_group_by_id(dict_['child'], id_)
                if result:
                    return result
        return {}

    @staticmethod
    def _norm_path(path):
        """
        Normalize a GROUP path.

        Parameters
        ----------
        path : str
            The path of a GROUP.

        Returns
        -------
        str
            The normalized path.
        """
        if not path.startswith('All marks'):
            path = 'All marks' if path == '' else f'All marks/{path}'
        return path

    @staticmethod
    def _rgb_to_int(rgb):
        """
        Represent a RGB tuple as a negative integer ranging from -1 to -16777216.

        Parameters
        ----------
        rgb : (int, int, int)
            A 3-tuple representing the RGB color system. The first element
            is for RED, the second is for GREEN and the third is for BLUE.
            Each digit ranges from 0 to 255.

        Returns
        -------
        int
            A negative integer for the given RGB value.

        Examples
        ---------
        >>> rgb_to_meta_color_id(255, 255, 255)
        >>> -1
        >>> rgb_to_meta_color_id(255, 255, 254)
        >>> -2
        >>> rgb_to_meta_color_id(255, 255, 0)
        >>> -256
        >>> rgb_to_meta_color_id(255, 254, 255)
        >>> -257

        Notes
        -----
        Colors in SIS are represented as RGB values in the UI, and as
        negative integers in the meta files.
        """
        global BASE_VALUE 
        global MAX_VALUE
        MAX_VALUE = 255
        BASE_VALUE = 256

        blue = MAX_VALUE - rgb[0]
        green = MAX_VALUE - rgb[1]
        red = MAX_VALUE - rgb[2]
        return -(blue*BASE_VALUE**2 + green*BASE_VALUE + red + 1)

    def get_id(self, path):
        """
        Get the ID of a GROUP.
        """
        path_ = self._norm_path(path).split('/')
        group = self._find_group(self.data, path_)

        if group is None:
            raise GroupNotFoundError(f'Cannot found the GROUP "{path}".')
        return group['id']

    def get_name(self, group_id):
        """
        Get the name of a GROUP.
        """
        group = self._find_group_by_id(self.data, group_id)
        return group.get('name', '')

    def is_path(self, path):
        """
        Check if the GROUP path exists.

        Parameter
        ---------
        path : str
            The path of a GROUP.

        Returns
        -------
        bool
            Return `True` if the GROUP path exists; otherwise, `False`.
        """
        path_list = self._norm_path(path).split('/')
        path_exists = self._find_group(self.data, path_list)

        if path_exists is None:
            return False
        return True

    def insert(self, path, color=None, hotkey=None, userparam=0):
        """
        Insert a GROUP.

        Parameters
        ----------
        path : str
            The path of the GROUP.
        color : (int, int, int) or None, default None
            A tuple representing the color in RGB system.
        hotkey : str or None, default None
            A single ASCII character that works as a shortcut when inserting marks.
            If ``None``, no shortcut is associated with the group.
        userparam : {0, 16, 32, 48}
            A set of values for making visible marks and texts.
            - 0: (visible_marks=True, text_visible=True)
            - 16: (visible_marks=True, text_visible=False)
            - 32: (visible_marks=False, text_visible=True)
            - 48: (visible_marks=False, text_visible=False)

        Examples
        --------
        Insert `MyCats` as a `GROUP`.

        >>> groups_segment = sis_meta.GroupsSegment()
        >>> groups_segment.insert('MyCats')
        
        Then, insert ``Akuma``, ``Kirris``, ``Lala``, ``Gato negro`` and
        ``Gato gordo`` as `SUBGROUPS` of `MyCats`.

        >>> groups_segment.insert('MyCats/Akuma')
        >>> groups_segment.insert('MyCats/Kirris')
        >>> groups_segment.insert('MyCats/Lala', (244, 219, 243), 49, 15)
        >>> groups_segment.insert('MyCats/Gato negro', color = (0, 0, 0))
        >>> groups_segment.insert('MyCats/Gato gordo', hotkey = 'r')
        """
        path_norm = self._norm_path(path)
        parents = path_norm.split('/')[:-1]
        name = path_norm.split('/')[-1]

        group = self._find_group(self.data, parents)

        if group is None:
            raise GroupNotFoundError(f'Cannot found the GROUP "{path}".')

        color_int = 0 if color is None else self._rgb_to_int(color)
        hotkey_int = 0 if hotkey is None else ord(hotkey.upper())

        self.group_id += 1
        dict_ = {
            'id': self.group_id,
            'name': name,
            'color': color_int,
            'hotkey': hotkey_int,
            'userparam': userparam,
        }
        group.setdefault('child', [])
        group['child'].append(dict_)

    def update(self, path, name=None, color=None, hotkey=None, userparam=None):
        """
        Update the attributes of a GROUP.

        path : str
            The path of the GROUP.
        name : str, default `None`
            The new name of the GROUP. If `None`, nothing happens.
        color : (int, int, int) or None, default None
            A tuple representing the color in RGB system.
        hotkey : int, default `None`. If `None`, nothing happens.
            A single ASCII character that works as a shortcut when inserting marks.
        userparam : {0, 16, 32, 48}
            A set of values for making visible marks and texts. See
            :meth:`~sis_meta.groups.GroupsSegment.insert`.
        """
        path = self._norm_path(path)
        path_list = path.split('/')

        group = self._find_group(self.data, path_list)
        if group is None:
            raise GroupNotFoundError(f'Cannot found {path} path.')

        if not name is None:
            group['name'] = name

        if not color is None:
            group['color'] = self._rgb_to_int(color)

        if not hotkey is None:
            group['hotkey'] = ord(hotkey)

        if not userparam is None:
            group['userparam'] = userparam

    def remove(self, path):
        """
        Remove a GROUP.

        Parameters
        ----------
        path : str
            The path of the target GROUP.
        """
        if not self.is_path(path):
            raise GroupNotFoundError(f'Cannot found the GROUP {path}.')

        path_list = self._norm_path(path).split('/')
        group_name = path_list[-1]
        group_path = path_list[:-1]

        group = self._find_group(self.data, group_path)
        subgroups = group['child']

        for index_, dict_ in enumerate(subgroups):
            if dict_['name'] == group_name:
                index = index_
                break
        subgroups.pop(index)

        if len(subgroups) == 0:
            group.pop('child')

class GroupNotFoundError(Exception):
    """
    Class for exception.
    """



"""
Read a meta file.
"""
from sis_meta.meta import Meta
from sis_meta.groups import GroupsSegment
from sis_meta.mark import GuideMark
from sis_meta.mark._parse import parse_marks_segment
from sis_meta.regex import META_PATTERN
from sis_meta.io.exceptions import MetaFileError

def read_from_file(path):
    """
    Read a meta file into a :class:`Meta`.

    Parameters
    ----------
    path : str or path-like object
        The path of meta file.

    Returns
    -------
    :class:`sis_meta.Meta`
        An object that contains marks.
    """
    with open(path, 'rb') as raw_file:
        raw_data = raw_file.read()

    match = META_PATTERN.match(raw_data)
    if not match:
        raise MetaFileError(f'Not a Meta File: {path}')

    meta = Meta() # Initialize a meta object

    # GROUPS_SEGMENT
    meta._groups = GroupsSegment(
        match.group('GROUPS_SEGMENT')
    )

    # MARKS_SEGMENT
    marks = parse_marks_segment(
        match.group('MARKS_SEGMENT')
    )
    list_ = []
    for position, length, text, group_id in marks:
        group_name = meta._groups.get_name(group_id)
        list_.append(
            GuideMark(position, length, text, group_id, group_name)
        )
    meta._data = list_
    return meta



"""
Serialize MARKS_GROUPS_DATA 
"""
from importlib.resources import files
import re

from sis_meta.mark._serialize import serialize_marks_segment
from sis_meta.io.exceptions import MetaFileError

MARKS_GROUPS_DATA = re.compile(
rb'''MARKS_GROUPS_DATA; BEGIN; COMPOSITE;BINARY;{{ TOTAL_SIZE }}(
VERSION; 2
.+
)

MARKS_GROUPS_DATA; END''', re.DOTALL
)

template_path = files('sis_meta.io') / 'templates/base.meta'

def write_meta(meta, path):
    """
    Write a :class:`sis_meta.Meta` object as a meta file.

    Parameters
    ----------
    meta : :class:`sis_meta.Meta`
        A :class:`sis_meta.Meta` object.
    path : path-like
        The path of the meta file.
    """
    if len(meta) == 0:
        raise MetaFileError('Cannot write a Meta object without marks!')

    content = serialize_marks_groups_data(meta)
    with open(path, 'wb') as meta_file:
        meta_file.write(content)

def serialize_marks_groups_data(meta):
    """
    Serialize a :class:`sis_meta.Meta` object.
    """
    groups_segment_bytes = str(meta._groups).encode()
    marks_segment_bytes = serialize_marks_segment(meta)

    with open(template_path, 'rb') as meta_file:
        content = meta_file.read()
        content = content.replace(b'{{ GROUPS_SEGMENT }}', groups_segment_bytes)
        content = content.replace(b'{{ MARKS_SEGMENT }}', marks_segment_bytes)
        match = MARKS_GROUPS_DATA.match(content)
        if match:
            size = str(len(match.group(1))).encode()
            content = content.replace(b'{{ TOTAL_SIZE }}', size)
    return content



"""
Save regular expressions.
"""
import re

META_PATTERN = re.compile(
rb'''(?P<META_HEADER>MARKS_GROUPS_DATA; BEGIN; COMPOSITE;BINARY;\d+
VERSION; 2)

(?P<GROUPS_SEGMENT>GROUPS_SEGMENT; BEGIN; COMPOSITE;BINARY;\d+
.+


GROUPS_SEGMENT; END)

(?P<MARKS_SEGMENT>MARKS_SEGMENT; BEGIN; COMPOSITE;BINARY;\d+
.+


MARKS_SEGMENT; END)


MARKS_GROUPS_DATA; END''', re.DOTALL
)



"""
Class for guide marks.
"""
class GuideMark:
    """
    A guide mark.
    """
    def __init__(self, position, length, text, group_id, group_name):
        self.position = position
        self.length = length
        self.text = text
        self.group_id = group_id
        self.group_name = group_name

    def __repr__(self):
        dict_ = {
            'position': round(self.position, 2),
            'length': round(self.length, 2),
            'text': self.text,
            'group_id': self.group_id,
            'group_name': self.group_name,
        }
        return dict_.__repr__()

    def is_interval(self):
        """
        Check if the guide mark is an interval.

        Returns
        -------
        bool : True or False
            `True` if the length is greater than 0. Otherwise, it returns `False`.
        """
        return self.length != 0

    def is_point(self):
        """
        Check if the guide mark is a point.

        Returns
        -------
        bool : True or False
            `True` if the length is greater than 0. Otherwise, it returns `False`.
        """
        return self.length == 0



"""
Parse MARKS_SEGMENT.
"""
import re
import struct

POSITIONS = re.compile(rb'''
POSITIONS; BEGIN; VECTOR_DOUBLE;BINARY;(\d+)
(.+)
POSITIONS; END''', re.DOTALL
)

IDS = re.compile(rb'''
IDS; BEGIN; VECTOR_INT;BINARY;(\d+)
(.+)
IDS; END''', re.DOTALL
)

LENGTHS = re.compile(rb'''
LENGTHS; BEGIN; VECTOR_DOUBLE;BINARY;(\d+)
(.+)
LENGTHS; END''', re.DOTALL
)

TEXTS_LONG = re.compile(rb'''
TEXTS; BEGIN; VECTOR_STRING
(.+)
TEXTS; END
''', re.DOTALL
)

TEXTS_SHORT = re.compile(rb'''
TEXTS; (.+)

TEXT_ATTR_POSITIONS; BEGIN;''', re.DOTALL
)

def parse_marks_segment(raw_data):
    """
    Parse MARKS_SEGMENT.

    Parameters
    ----------
    raw_data : bytes
        The data in a meta file. It MUST contain POSITIONS, IDS, LENGTHS
        and TEXTS tags.

    Returns
    -------
    zip of tuples, [(position, length, text, group_id, group_name), ...]
        A zip of tuples. Each tuple contains the start position,
        length, text, group_id and group_name of a mark.
    """
    # Read positions
    positions_match = POSITIONS.search(raw_data)
    #position_size = positions_match.group(1)
    positions_bytes = positions_match.group(2)[4:]
    positions = struct.unpack(f'{len(positions_bytes) // 8}d', positions_bytes)

    # Read ids / also known as groups
    ids_match = IDS.search(raw_data)
    #ids_size = ids_match.group(1)
    ids_bytes = ids_match.group(2)[4:]
    group_ids = struct.unpack(f'{len(ids_bytes) // 4}i', ids_bytes)

    # Read lengths
    lengths_match = LENGTHS.search(raw_data)
    #lengths_size = lengths_match.group(1)
    lengths_bytes = lengths_match.group(2)[4:]
    lengths = struct.unpack(f'{len(lengths_bytes) // 8}d', lengths_bytes)

    # Read texts
    texts_match = TEXTS_LONG.search(raw_data)
    if texts_match is None:
        texts_match = TEXTS_SHORT.search(raw_data)
    texts_bytes = texts_match.group(1)
    texts = []
    for text_bytes in texts_bytes.split(b';'):
        text_bytes = text_bytes.replace(bytes.fromhex('02'), b'\n') #Start of Text: U+0002
        text_bytes = text_bytes.replace(bytes.fromhex('03'), b';') #End of Text: U+0003
        text_bytes = text_bytes.replace(b'_|!!|_nuse', b'') #End of Text: U+0003
        texts.append(text_bytes.decode())
    return zip(positions, lengths, texts, group_ids)



"""
Parse GROUPS_SEGMENT into a list of dict. 
"""
import re
import io

GROUP = re.compile(rb'''ID; (?P<id>\d+)

NAME; (?P<name>.+)

COLOR; (?P<color>-{0,1}\d+)

HOTKEY; (?P<hotkey>\d+)

USERPARAM; (?P<userparam>\d+)

'''
)
GROUP_OPEN = re.compile(rb'GROUP(\d+); BEGIN; COMPOSITE;BINARY;(\d+)\n')
#GROUP_CLOSE = re.compile(rb'GROUP\d; END')
SUBGROUP_OPEN = re.compile(rb'SUBGROUPS; BEGIN; COMPOSITE;BINARY;\d+\n')
SUBGROUP_CLOSE = re.compile(rb'SUBGROUPS; END\n')

def parse_groups_segment(raw_data):
    """
    Discover group tree structure recursively.

    Parameters
    ----------
    raw_data : bytes
        The GROUPS_SEGMENT data in meta files.

    Returns
    -------
    list of dict
        Returns the GROUPS_SEGMENT in structured way.
    """
    data = io.BytesIO(raw_data)
    return discover_grouptree(data, [])

def discover_grouptree(data, tree):
    """
    Discover group tree structure recursively.

    Parameters
    ----------
    data : :class:`io.BytesIO`
        Binary stream containing the GROUPS_SEGMENT data in meta files.
    tree : list
        It contains the mark groups info.

    Returns
    -------
    list of dict
        Returns the GROUPS_SEGMENT info.
    """
    while True:
        current_position = data.tell()
        line = data.readline()

        # Base case
        if line == b'': # When riching the end of the document
            break

        if GROUP_OPEN.match(line):
            current_position = data.tell()
            data_temp = data.read()

            if GROUP.match(data_temp):
                group_content = GROUP.match(data_temp)
                dict_ = {
                    'id': int(group_content.group('id').decode()),
                    'name': group_content.group('name').decode(),
                    'color': group_content.group('color').decode(),
                    'hotkey': group_content.group('hotkey').decode(),
                    'userparam': group_content.group('userparam').decode(),
                }
                tree.append(dict_)
                current_position+= len(group_content.group(0))
            data.seek(current_position)

        # Recursive condition
        elif SUBGROUP_OPEN.match(line):
            tree[-1]['child'] = discover_grouptree(data, [])

        elif SUBGROUP_CLOSE.match(line):
            break

    return tree



"""
Serialize MARKS_SEGMENT 
"""
import re
from struct import pack
from importlib.resources import files

MAX_BYTES_TEXT_MARK = 1022
MAX_BYTES_ALL_TEXT_MARKS = 3001

MARKS_SEGMENT = re.compile(
    rb'MARKS_SEGMENT; BEGIN;.+?(\n.+)\n\nMARKS_SEGMENT; END', re.DOTALL
)

template_path = files('sis_meta.mark') / 'templates/marks_segment.meta'

def serialize_marks_segment(meta):
    """
    Serialize the marks in a :class:`meta_sis.Meta` object.

    Parameters
    ----------
    meta : :class:`meta_sis.Meta`
        A class for handling Meta files.

    Returns
    -------
    bytes
        The content of MARKS_SEGMENT
    """
    data = {
        'positions': [],
        'lengths': [],
        'texts': [],
        'ids': [],
    }

    for mark in meta:
        data['positions'].append(mark.position)
        data['lengths'].append(mark.length)
        data['ids'].append(mark.group_id)
        data['texts'].append(mark.text)

    meta = build_marks_segment(
        data['positions'],
        data['lengths'],
        data['ids'],
        data['texts']
    )
    return meta

def build_marks_segment(positions, lengths, ids, texts):
    """
    Build the content of MARKS_SEGMENT.

    Parameters
    -----------
    positions : list of float
        The starting point of the guide marks.
    lengths : list of float
        The duration of the guide marks.
    ids : list of int
        The group each guide mark belong to.
    texts : list of str
        The content of the guide marks.

    Returns
    -------
    bytes
        The content of a meta file.

    Notes
    -----
    The parameters must be equal in length. In the other hand, ids are 
    limited to the ones existing in the GROUPS_SEGMENT. By default they are:
        - 2 for `Single`
        - 3 for `Sounds`
        - 4 for `Noises`
        - 6 for `Speakers/M1`
        - 7 for `Speakers/M2`
        - 8 for `Speakers/F1`
        - 9 for `Speakers/F2`
        - 10 for `VAD`
        - 11 for `For_AutoCmp`
        - 13 for `Edit_Tracker/ET_LM`

    TEXT_ATTR_POSITIONS, TEXT_ATTR_LENGTHS, TEXT_ATTR_TYPES and
    TEXT_ATTR_VALUES seem to have the same attributes; so I decided to
    copy the values of TEXT_ATTR_POSITIONS on the other's parts.

    The lists pass as arguments must have the same lenght.
    """
    # Load template
    data = template_path.read_bytes()

    n_items = len(positions)

    # Replace values:
    # MARKS_SEGMENT
    # POSITIONS
    bytes_block = pack(f'=I{n_items}d', n_items, *positions)
    data = _update_meta(data, bytes_block, '{{ positions_size }}', '{{ positions }}')

    # LENGTHS
    bytes_block = pack(f'=I{n_items}d', n_items, *lengths)
    data = _update_meta(data, bytes_block, '{{ lengths_size }}', '{{ lengths }}')

    # IDS
    bytes_block = pack(f'=I{n_items}L', n_items, *ids)
    data = _update_meta(data, bytes_block, '{{ ids_size }}', '{{ ids }}')

    # TEXTS
    bytes_block = _norm_text(texts)
    if n_items < 16:
        bytes_block = b'TEXTS; ' + bytes_block
    else:
        bytes_block = b'TEXTS; BEGIN; VECTOR_STRING\n' + bytes_block + b'\nTEXTS; END'
    data = _update_meta(data, bytes_block, data_tag = '{{ texts }}')

    # TEXT_ATTR_POSITIONS
    zeroes = [0]*n_items
    bytes_block = pack(f'=I{n_items}I', n_items, *zeroes)
    data = _update_meta(data, bytes_block, '{{ attr_size }}', '{{ text_attr_positions }}')

    # TEXT_ATTR_LENGTHS

    # TEXT_ATTR_TYPES

    # TEXT_ATTR_VALUES

    # Calc sizes
    match = MARKS_SEGMENT.match(data)
    data = data.replace(b'{{ marks_segment_size }}', str(len(match.group(1))).encode())

    return data

def _norm_text(text_list):
    """
    Normalize texts in SIS format.

    Parameters
    ----------
    text_list : str
        A list of texts encoded in utf-8.

    Returns
    -------
    bytes
        A block of bytes containing the texts required in SIS format.

    Notes
    -----
    SIS does not allow some characters as texts:
        - Semicolons (``;``) are replaced by ``U+0002``
        - New lines (``\n``) are replaced by ``U+0003``
        - Empty characters (``''``) are replaced by ``_|!!|_nuse``
        - Tabs can be inserted in SIS; However, they are removed when they are
        read again. Characters at the rights are also removed.
        - It seems that each text mark is limited to 1023 bytes.
        - It seems that all text marks are limited to less than 3000 bytes.
    """
    list_ = []
    for text in text_list:
        text = text.encode()
        text = text.replace(b'\n', bytes.fromhex('02')) #Start of Text: U+0002
        text = text.replace(b';', bytes.fromhex('03')) #End of Text: U+0003
        text = text.replace(b'\t', b' ')
        if text == b'':
            text = b'_|!!|_nuse'

        if len(text) > MAX_BYTES_TEXT_MARK:
            text = text[0:MAX_BYTES_TEXT_MARK].decode(errors = 'ignore').encode()
        list_.append(text)

    bytes_block = b';'.join(list_)
    index = 0
    while len(bytes_block) > MAX_BYTES_ALL_TEXT_MARKS:
        index -= 1
        list_[index] = b'1'
        bytes_block = b';'.join(list_)
    return bytes_block

def _update_meta(data, bytes_block, size_tag = None, data_tag = None):
    """
    Update the content of the meta template.

    Parameters
    ----------
    data : bytes
        The content of the meta template.
    bytes_block : bytes
        The new data to be inserted in the tags.
    size_tag : str
        The name of the tag where the size of the data given bytes will be replaced
        with. The size is calculated automatically from the ``byte_blocks`` 
        parameter.
    data_tag : str
        The name of the tag where the ``byte_blocks`` will be replaced with.

    Returns
    -------
    bytes
        A new version of the ``data`` parameter where values in tags are 
        replaced with those in ``byte_blocks``.

    Notes
    -----
    These are the valid tags:
        {{ positions_size }}, {{ positions }}
        {{ ids_size }}, {{ ids }}
        {{ lengths_size }}, {{ lengths }}
        {{ attr_size }}, {{ text_attr_positions }}
    """
    if not size_tag is None:
        data = data.replace(size_tag.encode(), f'{len(bytes_block)}'.encode())
    if not data_tag is None:
        data = data.replace(data_tag.encode(), bytes_block)
    return data



"""
Handle annotation files (.meta).
"""
from sis_meta.groups import GroupsSegment
from sis_meta.mark import GuideMark
from sis_meta.io.write_meta import write_meta

class Meta:
    """
    Class for handling guide marks.
    """
    def __init__(self):
        self._data = []
        self._groups = GroupsSegment()

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def _sort(self):
        self._data = sorted(self._data, key=lambda x: x.position)

    def insert_group(self, group_path, *args, **kwargs):
        """
        Insert a new group.

        Parameters
        ----------
        group_path : str
            The path of the group.
        **args : tuple
            These parameters will be passed to 
            :meth:`~sis_meta.groups.GroupsSegment.insert`
        **kwargs : dict
            These parameters will be passed to 
            :meth:`~sis_meta.groups.GroupsSegment.insert`
        """
        self._groups.insert(group_path, *args, **kwargs)

    def update_group(self, group_path, new_name=None, **kwargs):
        """
        Update the attributes in a group.

        Parameters
        ----------
        group_path : str
            The path of the group.
        name : str or None, default None.
            The new name of the group.
        **kwargs : dict
            These parameters will be passed to 
            :meth:`~sis_meta.groups.GroupsSegment.update`
        """
        self._groups.update(group_path, new_name, **kwargs)
        if not new_name is None:
            # Update guide marks
            group_name = group_path.split('/')[-1]
            for mark in self:
                if not mark.group_name == group_name:
                    continue
                mark.group_name = new_name

    def remove_group(self, group_path):
        """
        Remove a group and the guide marks associated with it.

        Parameters
        ----------
        group_path : str
            The path of the group.
        """
        self._groups.remove(group_path)
        group_name = group_path.split('/')[-1]
        self._data = [mark for mark in self if not mark.group_name == group_name]

    def insert_guide_mark(self, group_name, position, length=0, text=''):
        """
        Insert a guide mark.
        """
        mark_group_id = self._groups.get_id(group_name)
        mark_group_name = group_name.split('/')[-1]
        guide_mark = GuideMark(
            position, length, text, mark_group_id, mark_group_name
        )
        self._data.append(guide_mark)
        self._sort()

    def write(self, path):
        """
        Write a meta file.
        """
        write_meta(self, path)



"""
Serialize GROUPS_SEGMENT.
"""
import re
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape

env = Environment(
    loader = PackageLoader("sis_meta.groups"),
    autoescape = select_autoescape()
)
env.trim_blocks = True
env.lstrip_blocks  = True
content_template = env.get_template('group.meta')
header_template = env.get_template('group_header.meta')

GROUP = re.compile(r'''GROUP\d+; BEGIN; COMPOSITE;BINARY;(
ID; \d+

NAME; .+

COLOR; -{0,1}\d+

HOTKEY; \d+

USERPARAM; \d+
)

GROUP\d+; END
'''
)

SUBGROUPS = re.compile(
r'''GROUP\d+; BEGIN; COMPOSITE;BINARY;
(.+
SUBGROUPS; BEGIN; COMPOSITE;BINARY;
(.+\n)
SUBGROUPS; END\n
)
GROUP\d+; END
''', re.DOTALL
)

GROUPS_SEGMENT = re.compile(
r'''GROUPS_SEGMENT; BEGIN; COMPOSITE;BINARY;
(.+\n)
GROUPS_SEGMENT; END''', re.DOTALL
)

def _serialize_groups_segment_content(data):
    """
    Serialize GROUPS_SEGMENT.

    Parameters
    ----------
    data : list of dict
        :class:`sis_meta.GroupsSegment`
    """
    group_str = ''
    for dict_ in data:
        child = dict_.get('child')
        group_size = None
        subgroup_size = None
        if child is None:
            temp_text = content_template.render(group = dict_)
            match = GROUP.match(temp_text)
            if match:
                group_size = len(match.group(1).encode())
            group_str += content_template.render(group= dict_, size= group_size)
        else:
            subgroup = _serialize_groups_segment_content(child)
            text = content_template.render(group = dict_, subgroup = subgroup)
            match = SUBGROUPS.match(text)
            if match:
                group_size = len(match.group(1).encode())
                subgroup_size = len(match.group(2).encode())
                group_size += len(str(subgroup_size).encode())

            subgroup_str = content_template.render(
                group = dict_,
                size = group_size,
                subgroup = subgroup,
                subgroup_size = subgroup_size
            )

            group_str += subgroup_str
    return group_str

def serialize_groups_segment(data):
    """
    Serialize GROUPS_SEGMENT.

    Parameters
    ----------
    data : list of dict
        :class:`sis_meta.GroupsSegment`
    """
    content = _serialize_groups_segment_content(data)
    text = header_template.render(content = content)
    match = GROUPS_SEGMENT.match(text)
    if match:
        groups_segment_size = len(match.group(1).encode())
    return header_template.render(content = content, size= groups_segment_size)
